-----------------------------------------------------------------------------------------
-- Start Activity
-----------------------------------------------------------------------------------------
function TowerAssault:StartActivity()
	print ("TowerAssault:StartActivity");

	if self.IsInitialized == nil then
		self.IsInitialized = false
	end

	if self.IsInitialized then 
		return
	end

	self.ModuleName = "TowerAssault.rte";
	
	self.IsInitialized = true
	
	self.PlayerTeam = 0
	self.CPUTeam = 1

	self.PlayerFaction = self:GetTeamTech(self.PlayerTeam)
	self.CPUFaction = self:GetTeamTech(self.CPUTeam)

	self.LastPlayerSpawnTime = 0
	self.LastEnemySpawnTime = 0

	self.Time = 0;

	self.PlayerTroops = 0
	self.EnemyTroops = 0

	self.TickInterval = 1000
	
	self.SentryProbability = 0.025
	
	-- Default settings
	self.SendPlayerTroopsToUpperLevel = true
	
	self.WeaponsPerUnit = 1;
	
	self.PointsToCapture = 850
	self.PointsPerUnit = 10
	self.PointsPerBrain = 75

	self.PlayerBudget = self:GetStartingGold();
	
	if self:GetFogOfWarEnabled() then
		SceneMan:MakeAllUnseen(Vector(96,96), self.PlayerTeam)
	end
	
	self.MaxPlayerTroops = 8
	self.MaxEnemyTroops = 12
	
	self.PlayerSpawnInterval = 10
	self.EnemySpawnInterval = 15
	
	self.PlayerIncome = 5
	
	self.MultiSpawnProbablity = 0.085
	self.MultiSpawnAmount = 5
	
	-- Default weapon types
	self.PlayerWeapons = {"Light Weapons", "Light Weapons", "Heavy Weapons", "Sniper Weapons"}
	self.PlayerActors = {"Light Infantry", "Light Infantry", "Light Infantry", "Heavy Infantry"}
	
	self.CPUWeapons = {"Light Weapons", "Light Weapons", "Heavy Weapons", "Sniper Weapons"}
	self.CPUActors = {"Light Infantry", "Light Infantry", "Light Infantry", "Heavy Infantry"}
	
	-- Difficulty-specific values
	if self.Difficulty <= GameActivity.CAKEDIFFICULTY then
		self.PositiveIndex = 1.3;
		self.NegativeIndex = 0.7;
	elseif self.Difficulty <= GameActivity.EASYDIFFICULTY then
		self.PositiveIndex = 1.15;
		self.NegativeIndex = 0.85;
	elseif self.Difficulty <= GameActivity.MEDIUMDIFFICULTY then
		self.PositiveIndex = 1.0;
		self.NegativeIndex = 1.0;
	elseif self.Difficulty <= GameActivity.HARDDIFFICULTY then
		self.PositiveIndex = 0.85;
		self.NegativeIndex = 1.15;
	elseif self.Difficulty <= GameActivity.NUTSDIFFICULTY then
		self.PositiveIndex = 0.70;
		self.NegativeIndex = 1.3;
	elseif self.Difficulty <= GameActivity.MAXDIFFICULTY then
		self.PositiveIndex = 0.55;
		self.NegativeIndex = 1.45;
	end			

	self.PlayerBudget = math.floor(self.PlayerBudget * self.PositiveIndex)
	
	self.MaxPlayerTroops = math.floor(self.MaxPlayerTroops * self.PositiveIndex) + (self.PlayerCount - 1)
	self.MaxEnemyTroops = math.floor(self.MaxEnemyTroops * self.NegativeIndex)
	
	self.PlayerSpawnInterval = math.floor(self.PlayerSpawnInterval * self.NegativeIndex)
	self.EnemySpawnInterval = math.floor(self.EnemySpawnInterval * self.PositiveIndex)
	
	self.PlayerIncome = math.floor(self.PlayerIncome * self.PositiveIndex)
	
	self.PointsToCapture = math.floor(self.PointsToCapture * self.NegativeIndex)

	
	print("Budget: "..self.PlayerBudget)
	print("Player income: "..self.PlayerIncome)
	print("Max player troops: "..self.MaxPlayerTroops)
	print("Max CPU troops: "..self.MaxEnemyTroops)
	print("Player spawn: "..self.PlayerSpawnInterval)
	print("CPU spawn: "..self.EnemySpawnInterval)
	print("Points to capture: "..self.PointsToCapture)
	
	self.PlayerBrainDead = true
	
	self.BrainPrice = 0
	
	local side = ""
	
	if MODE_ASCENDING == nil then
		MODE_ASCENDING = false
	end
	
	MODE_ASCENDING = not MODE_ASCENDING
	
	-- Get areas
	if MODE_ASCENDING == false then
		self.AtkBrain = SceneMan.Scene:GetArea("AtkBrain2")
		side = "2-"
	else
		self.AtkBrain = SceneMan.Scene:GetArea("AtkBrain")
	end

	-- Spawn player brain
	for p = 0, self.PlayerCount - 1 do
		local brain = RandomAHuman("Brains", self.PlayerFaction)
		if brain ~= nil then
			self.BrainPrice = brain:GetTotalValue(0, 1)

			brain.Team = self.PlayerTeam
			brain.Pos = self.AtkBrain:GetCenterPoint();
		
			for i = 1, self.WeaponsPerUnit do
				-- Give brain some weapons
				local weap = self:MakeAnyOfThoseWeapons(self.PlayerFaction, self.PlayerWeapons);
			
				if weap ~= nil then
					brain:AddInventoryItem(weap)
				end
			end
			MovableMan:AddActor(brain);
		end
	end
	
	-- Spawn attackers
	for i = 1, 10 do
		if SceneMan.Scene:HasArea("AtkTrooper"..side..i)then
			local act = self:MakeRandomActor(self.PlayerFaction, self.PlayerTeam, false, false, self.PlayerActors, self.PlayerWeapons);
			if act ~= nil then
				act.Pos = SceneMan.Scene:GetArea("AtkTrooper"..side..i):GetCenterPoint()
				act.AIMode = Actor.AIMODE_SENTRY
				MovableMan:AddActor(act)
			end
		else
			break;
		end
	end
	
	-- Spawn defenders
	for i = 1, 10 do
		if SceneMan.Scene:HasArea("DefTrooper"..side..i)then
			local act = self:MakeRandomActor(self.CPUFaction, self.CPUTeam, true, false, self.CPUActors, self.CPUWeapons);
			if act ~= nil then
				act.Pos = SceneMan.Scene:GetArea("DefTrooper"..side..i):GetCenterPoint()
				act.AIMode = Actor.AIMODE_SENTRY
				MovableMan:AddActor(act)
			end
		else
			break;
		end
	end
	
	-- Init levels
	-- First count available levels
	local maxlevels = 0
	
	for i = 1, 20 do
		if SceneMan.Scene:HasArea("Level"..i)then
			maxlevels = maxlevels + 1
		else
			break
		end
	end

	local lvl_from = 1
	local lvl_to = maxlevels
	local lvl_step = 1
	
	if MODE_ASCENDING == false then
		lvl_from = maxlevels
		lvl_to = 1
		lvl_step = -1
	end
	
	self.Levels = {}
	for i = lvl_from, lvl_to, lvl_step do
		local l = #self.Levels + 1

		self.Levels[l] = {}
		self.Levels[l]["LevelArea"] = SceneMan.Scene:GetArea("Level"..i)
		self.Levels[l]["SpawnPos"] = SceneMan.Scene:GetArea("Spawn"..i):GetCenterPoint()
		self.Levels[l]["Owner"] = self.CPUTeam
		self.Levels[l]["Points"] = 0
		self.Levels[l]["PointsInc"] = 0
		
		self.Levels[l]["Waypoints"] = {}
		local waypoints = {}
		
		for j = 1, 20 do
			if SceneMan.Scene:HasArea("Waypoint"..l.."-"..j)then
--				print ("Waypoint"..l.."-"..j)
				waypoints[j] = SceneMan.Scene:GetArea("Waypoint"..l.."-"..j):GetCenterPoint();
			else
				break;
			end
		end
		
		-- Reorder weapoints if we're descending
		if MODE_ASCENDING == false then
			local k = 1
		
			for j = #waypoints, 1 , -1 do
				self.Levels[l]["Waypoints"][k] = waypoints[j]
				k = k + 1
			end
		else
			self.Levels[l]["Waypoints"] = waypoints
		end
	end
	
	self.Levels[1]["Owner"] = self.PlayerTeam
	self.Levels[1]["Points"] = self.PointsToCapture
	self.Levels[1]["PointsInc"] = 0
	
	self.GenericTimer = Timer();
	self.GenericTimer:Reset();
	
	self.IsInitialized = true
	
	self:SetTeamFunds(0, 0);
end
-----------------------------------------------------------------------------------------
-- Pause Activity
-----------------------------------------------------------------------------------------
function TowerAssault:PauseActivity(pause)
end
-----------------------------------------------------------------------------------------
-- End Activity
-----------------------------------------------------------------------------------------
function TowerAssault:EndActivity()
end
-----------------------------------------------------------------------------------------
--
-----------------------------------------------------------------------------------------
function TowerAssault:MakeAnyOfThoseWeapons(faction, list)
	local r = math.random(#list)
	return RandomHDFirearm(list[r], faction)
end
-----------------------------------------------------------------------------------------
--
-----------------------------------------------------------------------------------------
function TowerAssault:GetModuleName(s)
	local pos;
	
	pos = string.find(s ,"/");
	
	if pos ~= nil then
		local  preset , module;

		module = string.sub(s , 1 , pos - 1);
		preset = string.sub(s , pos + 1 , string.len(s));

		return module
	else
		return nil
	end
end
-----------------------------------------------------------------------------------------
--
-----------------------------------------------------------------------------------------
function TowerAssault:MakeRandomActor(f, t, aiusableonly, pay, al, wl)
	local act = nil;
	local allowed = true

	if pay and t == self.PlayerTeam and self.PlayerBudget > 0 then
		allowed = false;
	end

	if allowed then
		local ar = math.random(#al)
	
		act = RandomAHuman(al[ar], f)
		if act ~= nil then
			local mdl = self:GetModuleName(act:GetModuleAndPresetName())

			act.Team = t

			-- Give weapons
			for i = 1, self.WeaponsPerUnit do
				local weap = self:MakeAnyOfThoseWeapons(mdl, wl)
				if weap ~= nil then
					act:AddInventoryItem(weap)
				end
			end
			
			if pay and t == self.PlayerTeam then 
				self.PlayerBudget = self.PlayerBudget - act:GetTotalValue(0, 1)
			end
		end
	end
	
	return act;
end
-----------------------------------------------------------------------------------------
--
-----------------------------------------------------------------------------------------
function TowerAssault:GetLevelStatistics()
	self.PlayerTroops = 0
	self.EnemyTroops = 0

	-- Clear old stats
	for i = 1, #self.Levels do
		self.Levels[i]["PlayerActors"] = 0
		self.Levels[i]["EnemyActors"] = 0
		self.Levels[i]["PlayerBrain"] = false
	end
	
	for actor in MovableMan.Actors do
		if actor.Pos.X < -50 or actor.Pos.X > SceneMan.SceneWidth + 50 or actor.Pos.Y > SceneMan.SceneHeight then
			actor.ToDelete = true;
		end
	
		-- Count troops
		if actor.Team == self.PlayerTeam then
			self.PlayerTroops = self.PlayerTroops + 1
		else
			self.EnemyTroops = self.EnemyTroops + 1
		end
	
		for i = 1, #self.Levels do
			if actor.Team == self.PlayerTeam then
				if self.Levels[i]["LevelArea"]:IsInside(actor.Pos) then
					if actor:IsInGroup("Brains") then
						self.Levels[i]["PlayerBrain"] = true
						break
					else
						self.Levels[i]["PlayerActors"] = self.Levels[i]["PlayerActors"] + 1
						break
					end
				end
			elseif actor.Team == self.CPUTeam then
				if self.Levels[i]["LevelArea"]:IsInside(actor.Pos) then
					self.Levels[i]["EnemyActors"] = self.Levels[i]["EnemyActors"] + 1
					
					-- Set forgotten actors to brainhunt
					if actor.AIMode == Actor.AIMODE_SENTRY and self.Levels[i]["Owner"] ~= self.CPUTeam then
						actor.AIMode = Actor.AIMODE_BRAINHUNT
					end
					break
				end
			end
		end
	end
	
	--for i = 1, #self.Levels do
	--	print(tostring(i).." - "..self.Levels[i]["PlayerActors"])
	--end
end
-----------------------------------------------------------------------------------------
--
-----------------------------------------------------------------------------------------
function TowerAssault:ShowOwners()
	for i = 1, #self.Levels do
		local text = ""
		
		-- Show remaining spawn time
		if self.Levels[i]["Owner"] == self.PlayerTeam and self.Levels[i]["PlayerBrain"] then
			text = text .. "Budget: "..self.PlayerBudget.."oz\n"

			if self.PlayerTroops < self.MaxPlayerTroops then
				text = text .. "Reinforcements in "..self.Time - (self.LastPlayerSpawnTime + self.PlayerSpawnInterval)
				text = text .. "\nUnits: "..self.PlayerTroops.." / "..self.MaxPlayerTroops;
			else
				text = text .. "Too many units!"
			end
		end
		
		-- Show capture progress
		local toshow = false;
		
		if (self.Levels[i]["PointsInc"] > 0 and self.Levels[i]["Owner"] == self.CPUTeam) then
			toshow = true
		end

		if (self.Levels[i]["PointsInc"] < 0 and self.Levels[i]["Owner"] == self.PlayerTeam) then
			toshow = true
		end
		
		if toshow then
			local pts = math.ceil(self.Levels[i]["Points"] / self.PointsToCapture * 100)
			local inc = math.ceil(self.Levels[i]["PointsInc"] / self.PointsToCapture * 100)
			
			local plus = ""
			
			if inc > 0 then
				plus = "+"
			end
			
			text = text .. "\n"..pts.."% [ "..plus..inc.."% ]"
			--text = text .. "\n"..self.Levels[i]["Points"].." [ "..plus..self.Levels[i]["PointsInc"].." ]"
			
			-- Show graphical progress
			redcount = math.floor((self.Levels[i]["Points"] / self.PointsToCapture * 100) / 20)
			
			for j = 0, 4 do
				local preset = "EmptyLevel";

				if j < redcount then
					preset = "PlayerLevel"
				else
					preset = "EnemyLevel"
				end
			
				local pix = CreateMOPixel(preset);
				pix.Pos = self.Levels[i]["SpawnPos"] + Vector(0,-24 - (j * 12)) 
				MovableMan:AddParticle(pix);
			end
		else
			local preset = "EmptyLevel";
		
			if self.Levels[i]["Owner"] == self.PlayerTeam then
				preset = "PlayerLevel"
			elseif self.Levels[i]["Owner"] == self.CPUTeam then
				preset = "EnemyLevel"
			end
			
			local pix = CreateMOPixel(preset);
			pix.Pos = self.Levels[i]["SpawnPos"] + Vector(0,-40)
			MovableMan:AddParticle(pix);
		end
		
		if string.len(text) > 0 then
			self:AddObjectivePoint(text, self.Levels[i]["SpawnPos"] + Vector(0,-120) , self.PlayerTeam, GameActivity.ARROWDOWN);
		end
	end
end
-----------------------------------------------------------------------------------------
-- Update Activity
-----------------------------------------------------------------------------------------
function TowerAssault:SpawnReinforcements(team)
	local actors; 
	local limit;
	
	if team == self.PlayerTeam then
		actors = self.PlayerTroops;
		limit = self.MaxPlayerTroops;
	elseif team == self.CPUTeam then
		actors = self.EnemyTroops;
		limit = self.MaxEnemyTroops;
	end
	
	if actors >= limit then
		return
	end

	local level = nil;
	-- Find a suitable level to spawn reinforcements
	if team == self.PlayerTeam then
		for i = 1, #self.Levels do
			if self.Levels[i]["Owner"] == self.PlayerTeam and self.Levels[i]["PlayerBrain"] then
				level = i
			end
		end
	elseif team == self.CPUTeam then
		-- Check levels from top and spawn on level adjacent to upper AI levels
		-- Stop if we've got enemy units in our area
		for i = #self.Levels, 1, -1 do
			if self.Levels[i]["Owner"] == self.CPUTeam then
				level = i
				
				if self.Levels[i]["PlayerBrain"] or self.Levels[i]["PlayerActors"] > 0 then
					break;
				end
			else
				break
			end
		end
	end
	
	-- Finally spawn something
	if level ~= nil then
		-- Spawn player troops
		if team == self.PlayerTeam then
			local act = self:MakeRandomActor(self.PlayerFaction, self.PlayerTeam, false, false, self.PlayerActors, self.PlayerWeapons);
			if act ~= nil then
				act.Pos = self.Levels[level]["SpawnPos"]
				if self.SendPlayerTroopsToUpperLevel then
					-- Find level where we need to go, thats the first enemy level above ours
					for i = level + 1, #self.Levels do
						if self.Levels[i]["Owner"] ~= self.PlayerTeam then
							act.AIMode = Actor.AIMODE_GOTO
							act:ClearAIWaypoints();
							
							-- If we have some waypoints then use them
							for j = 1, #self.Levels[level]["Waypoints"] do
								act:AddAISceneWaypoint(self.Levels[level]["Waypoints"][j])
							end
							
							act:AddAISceneWaypoint(self.Levels[i]["SpawnPos"])
							break;
						end
					end
				else
					act.AIMode = Actor.AIMODE_SENTRY
				end
				MovableMan:AddActor(act)
			end
		elseif team == self.CPUTeam then
			-- Spawn CPU troops
			local act = self:MakeRandomActor(self.CPUFaction, self.CPUTeam, true, false, self.CPUActors, self.CPUWeapons);
			if act ~= nil then
				act.Pos = self.Levels[level]["SpawnPos"]
				
				if math.random() < self.SentryProbability then
					act.AIMode = Actor.AIMODE_SENTRY
				else
					act.AIMode = Actor.AIMODE_BRAINHUNT
				end
				--print ("Spawn enemy")
				MovableMan:AddActor(act)
			end
		end
	end
end
-----------------------------------------------------------------------------------------
-- Update Activity
-----------------------------------------------------------------------------------------
function TowerAssault:ApplyPoints()
	for i = 1, #self.Levels do
		local points = self.Levels[i]["PlayerActors"] * self.PointsPerUnit - self.Levels[i]["EnemyActors"] * self.PointsPerUnit;
		if self.Levels[i]["PlayerBrain"] then
			points = points + self.PointsPerBrain;
		end
		
		self.Levels[i]["Points"] = self.Levels[i]["Points"] + points;
		self.Levels[i]["PointsInc"] = points
		
		if self.Levels[i]["Points"] <= 0 then
			self.Levels[i]["Points"] = 0;
			
			if self.Levels[i]["Owner"] ~= self.CPUTeam then
				self.Levels[i]["Owner"] = self.CPUTeam
			end
		end
		
		if self.Levels[i]["Points"] >= self.PointsToCapture then
			self.Levels[i]["Points"] = self.PointsToCapture;
			
			if self.Levels[i]["Owner"] ~= self.PlayerTeam then
				self.Levels[i]["Owner"] = self.PlayerTeam
				
				-- Send all unused troops to upper level
				if self.SendPlayerTroopsToUpperLevel then
					for j = i + 1, #self.Levels do
						if self.Levels[j]["Owner"] ~= self.PlayerTeam then
							for actor in MovableMan.Actors do
								if actor.Team == self.PlayerTeam and not actor:IsInGroup("Brains") then
									actor.AIMode = Actor.AIMODE_GOTO
									actor:ClearAIWaypoints();
									actor:AddAISceneWaypoint(self.Levels[j]["SpawnPos"])
								end
							end
							break;
						end
					end
				end
				
				-- Mark CPU levels as cut off
				--[[for j = i, 1, -1 do
					if self.Levels[j]["Owner"] == self.CPUTeam then
						self.Levels[j]["Owner"]	= -1
						self.Levels[j]["Points"] = self.PointsToCapture / 2
					end
				end]]--
			end
		end
	end
	
	-- Check winning conditions
	if self.Levels[#self.Levels]["Owner"] == self.PlayerTeam then
		-- Victory
		if self.ActivityState ~= Activity.OVER then
			self.WinnerTeam = self.PlayerTeam;
			ActivityMan:EndActivity();
		end
	end
end
-----------------------------------------------------------------------------------------
-- Update Activity
-----------------------------------------------------------------------------------------
function TowerAssault:UpdateActivity()
	-- Just check for intialization flags in update loop to avoid unnecessary function calls during all the mission
	if self.IsInitialized == nil then
		self.IsInitialized = false
	end
	
	if not self.IsInitialized then
		--Init mission if we're still not
		self:StartActivity()
	end

	self:ClearObjectivePoints();
	
	self:GetLevelStatistics();
	self:ShowOwners()
	
	if self.ActivityState ~= Activity.OVER then
		if self.GenericTimer:IsPastSimMS(self.TickInterval) then
			self.Time = self.Time + 1;
			self.GenericTimer:Reset();
			
			if self.Time >= self.LastPlayerSpawnTime + self.PlayerSpawnInterval then
				self:SpawnReinforcements(self.PlayerTeam)
				self.LastPlayerSpawnTime = self.Time;
			end

			if self.Time >= self.LastEnemySpawnTime + self.EnemySpawnInterval then
				local count = 1
				
				if math.random() < (self.MultiSpawnProbablity * 2) then
					count = math.ceil(self.MultiSpawnAmount / 2)
				else
					if math.random() < self.MultiSpawnProbablity then
						count = self.MultiSpawnAmount
					end
				end
			
				for i = 1, count do
					self:SpawnReinforcements(self.CPUTeam)
				end
				self.LastEnemySpawnTime = self.Time;
			end
			
			self:ApplyPoints();
			
			self.PlayerBudget = self.PlayerBudget + self.PlayerIncome
		end
	end
	
	self:DoBrainSelection()
	
	self:YSortObjectivePoints();--]]--
end
-----------------------------------------------------------------------------------------
-- Brain selection and gameover conditions check
-----------------------------------------------------------------------------------------
function TowerAssault:RespawnPlayerBrain()
	-- Find suitable level
	local lastcleanlevel
	local lastplayerlevel
	
	for i = 1, #self.Levels do
		if self.Levels[i]["Owner"] == self.PlayerTeam then
			lastplayerlevel = i
			if self.Levels[i]["EnemyActors"] == 0 then
				lastcleanlevel = i
			end
		end
	end
	
	local level;
	
	if lastcleanlevel ~= nil then
		level = lastcleanlevel
	else
		level = lastplayerlevel
	end
	
	--print (lastcleanlevel)
	--print (lastplayerlevel)

	if level ~= nil and self.PlayerBudget >= self.BrainPrice then
		local brain = RandomAHuman("Brains", self.PlayerFaction)
		if brain ~= nil then
			self.PlayerBudget = self.PlayerBudget - self.BrainPrice

			brain.Team = self.PlayerTeam
			brain.Pos = self.Levels[level]["SpawnPos"]
		
			for i = 1, self.WeaponsPerUnit do
				-- Give brain some weapons
				local weap = self:MakeAnyOfThoseWeapons(self.PlayerFaction, self.PlayerWeapons, false);
				
				if weap ~= nil then
					brain:AddInventoryItem(weap)
				end
			end
			MovableMan:AddActor(brain);
			return brain;
		end
	end
	
	return nil;
end
-----------------------------------------------------------------------------------------
-- Brain selection and gameover conditions check
-----------------------------------------------------------------------------------------
function TowerAssault:DoBrainSelection()
	if self.ActivityState ~= Activity.OVER then
		for player = 0, self.PlayerCount - 1 do
			local team = self:GetTeamOfPlayer(player);
			local brain = self:GetPlayerBrain(player);

			if not brain or not MovableMan:IsActor(brain) or not brain:HasObjectInGroup("Brains") then
				if team == self.PlayerTeam then
					self.PlayerBrainDead = true
				end

				self:SetPlayerBrain(nil, player);
				local newBrain = MovableMan:GetUnassignedBrain(team);
				
				if not newBrain then
					newBrain = self:RespawnPlayerBrain();
				end				
				
				if newBrain then
					self:SetPlayerBrain(newBrain, player);
					self:SwitchToActor(newBrain, player, team);
					
					if team == self.PlayerTeam then
						self.PlayerBrainDead = false
					end
				end
			else
				self:AddObjectivePoint("Protect!!!", brain.AboveHUDPos , self.PlayerTeam, GameActivity.ARROWDOWN);
				self:SetObservationTarget(brain.Pos, player)
			end
		end
	end
end
-----------------------------------------------------------------------------------------
-- Thats all folks!!!
-----------------------------------------------------------------------------------------